<div ng-controller="UploadCtrl">
  <div class="upload col-xs-12 col-md-offset-3 col-md-6" collapse="uploadCollapsed">
    <form role="form" ng-submit="submit()">
      <div class="row">
        <div class="col-xs-6 text-left">
          <input ng-show="uploader.queue.length === 0" ng-file-select type="file" />
          <span ng-show="uploader.queue.length > 0" >
            <button type="button" class="close" aria-hidden="true" style="float: left; margin-right: 10px" ng-click="uploader.clearQueue()">&times;</button>
            {{ uploader.queue[0].file.name }}
          </span>
        </div>
        <div class="col-xs-6 text-right">
          <button ng-disabled="!canUpload()" ng-click="uploadPackage()" class="btn btn-success">
            <i class="fa fa-refresh fa-spin" ng-show="uploading"></i>
            <i class="fa fa-upload" ng-hide="uploading"></i>
          </button>
        </div>
      </div>
    </form>
  </div>
</div>
